import os

try:
    from math import exp, isclose
except:
    from math import exp
    from numpy import isclose

from whiscy.modules.pam_data import code, logpameigval, pameigvec, pameigvecinv


class Distance:
    def __init__(self, seq=0, dist=0.0, mat=None, expect=None):
        self.seq = seq
        self.dist = dist
        if mat is None:
            self.mat = [[0 for x in range(20)] for y in range(20)]
        else:
            self.mat = mat
        if expect is None:
            self.expect = [0.0 for x in range(20)]
        else:
            self.expect = expect


def get_pam_assemble(distance):
    m = [[0.0 for x in range(20)] for y in range(20)]
    disteigval = [0.0 for x in range(20)]

    for n in range(20):
        disteigval[n] = exp(distance * logpameigval[n])

    for j in range(20):
        for i in range(20):
            v = 0
            for n in range(20):
                v += pameigvec[i][n] * disteigval[n] * pameigvecinv[n][j]
            m[i][j] = v
    return m


def pam_load_sequences(alignment_file, distance_file):
    """Loads the aligment and distance files"""
    if not os.path.exists(distance_file):
        raise Exception("Distance file {0} does not exist".format(distance_file))

    if not os.path.exists(alignment_file):
        raise Exception("Sequence file {0} does not exist".format(alignment_file))

    refseq = ""
    seqtodis = []

    seqnr = 0
    distances = []
    with open(distance_file, "r") as input_distances:
        first_line = input_distances.readline().rstrip(os.linesep)
        fields = first_line.split()
        try:
            seqnr = int(fields[0])
        except:
            raise Exception("Can not read sequence number")
        if seqnr < 1 or seqnr > 10000:
            raise Exception("Invalid number of sequences")

        raw = input_distances.readline().rstrip(os.linesep).split()

        if len(raw) != (seqnr + 1):
            raise Exception("Reading error in distance file {0}".format(distance_file))

        for n in range(seqnr):
            seq = n
            dist = 0
            try:
                val = float(raw[1 + n])
                if val < 0 or val > 10:
                    raise ValueError()
                dist = val
            except ValueError:
                raise Exception(
                    "Reading error in distance file {0}".format(distance_file)
                )

            m = get_pam_assemble(100 * dist)

            expect = [0 for i in range(20)]
            for i in range(20):
                for ii in range(20):
                    expect[i] += m[i][ii] * m[i][ii]

            d = Distance(seq, dist, m, expect)
            distances.append(d)

    seqlen = 0
    sequences = [[] for _ in range(seqnr)]
    with open(alignment_file, "r") as input_alignment:
        first_line = input_alignment.readline().rstrip(os.linesep)
        fields = first_line.split()
        seqlen = int(fields[1])
        if seqlen < 1 or seqnr > 10000:
            raise Exception("Invalid sequence length")

        for n in range(seqnr):
            sequences[n] = []
            line = input_alignment.readline().rstrip(os.linesep)
            if line:
                name, sequence = line[:10], line[10:]
                if n == 0:
                    refseq = sequence
                for c in sequence:
                    sequences[n].append(code[ord(c)])

    # Sorted in ascending order as the C++ qsort
    sorted_distances = sorted(
        distances, key=lambda distance: distance.dist, reverse=False
    )
    seqtodis = [0 for _ in range(seqnr)]
    for n in range(seqnr):
        seqtodis[sorted_distances[n].seq] = n

    # print("***Dis***")
    # for x in distances:
    #     print("%.6f" % x.dist)
    #     print(' '.join([("%.6f" % i)  for i in x.expect]) + ' ')
    # print("******")

    # print("***SortedDis***")
    # for x in sorted_distances:
    #     print("%.6f" % x.dist)
    # print("******")

    # print("***Seqtodis***")
    # for x in seqtodis:
    #     print(x)
    # print("******")

    return seqnr, seqlen, refseq, sorted_distances, sequences, seqtodis


def pam_calc_similarity(pos, seqnr, seq, dis):
    nextnr = 0
    currnr = 0
    nextdist = 0.0
    currdist = 0.0
    lastdist = 0.0
    scores = [0.0 for _ in range(seqnr)]
    distances = [0.0 for _ in range(seqnr)]
    for n in range(1, seqnr):
        if seq[dis[n].seq][pos] >= 0:
            nextnr = n
            nextdist = dis[n].dist
            break
    if n == seqnr:
        return 0, distances, scores

    sim = 0.0
    totsim = 0.0
    weight = 0.5 * nextdist
    totweight = weight

    vref = seq[0][pos]
    counter = 0

    while True:
        lastdist = currdist
        currnr = nextnr
        currdist = nextdist
        for n in range(currnr + 1, seqnr):
            if seq[dis[n].seq][pos] >= 0:
                nextnr = n
                nextdist = dis[n].dist
                break
        if n == (seqnr - 1):
            break
        if isclose(currdist, lastdist):
            continue

        m = dis[currnr].mat
        vcomp = seq[dis[currnr].seq][pos]
        weight = 0.5 * (nextdist - lastdist)
        # This scaling factor of 2.4 is totally arbitrary, but gives a nice range of scores.
        # Scaling does not affect the final ranking of scores whatsoever
        sim = 2.4 * (m[vref][vcomp] - dis[currnr].expect[vref])

        totsim += weight * sim
        distances[counter] = currdist
        scores[counter] = totsim
        counter += 1

    return counter, distances, scores



import argparse
import logging
import math
import os
import sys
from pathlib import Path

from whiscy.modules import PARAM_PATH
from whiscy.modules.whiscy_data import (
    load_cons_file,
    load_residue_distance_matrix,
    load_smoothing_parameter_file,
)

logger = logging.getLogger("parasmooth")
logger.setLevel(logging.INFO)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(name)s [%(levelname)s] %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)


SMOOTHING_PARAM_FILE = Path(PARAM_PATH, "parasmooth.par")


def get_weight(par, val):
    """Calculates the weight in the par dictionary of the key val"""
    i = 0
    for k in sorted(par.keys()):
        if k > val:
            vh, wh = k, par[k]
            break
        i += 1

    if math.isclose(val, vh) or i == 0:
        return wh
    else:
        i -= 1
        key = list(par.keys())[i]
        vl = key
        wl = par[key]

        return wl + (wh - wl) * (val - vl) / (vh - vl)


def calculate_parasmooth(res_sur, res_lac, resdist, par):
    """Main function, calculates parasmooth values"""

    # Get the max distance
    maxdis = sorted(par.keys())[-1]

    for n in range(len(res_sur)):
        weight = 1.0
        score = res_sur[n].score
        for i in range(len(resdist)):
            if resdist[i].dis > maxdis:
                continue
            partner = -1

            if resdist[i].nr1 == res_sur[n].nr:
                partner = resdist[i].nr2
            elif resdist[i].nr2 == res_sur[n].nr:
                partner = resdist[i].nr1
            if partner == -1:
                continue

            found = False
            if resdist[i].dis < maxdis:
                for nn in range(len(res_sur)):
                    if n == nn:
                        continue
                    if res_sur[nn].nr == partner:
                        found = True
                        currweight = get_weight(par, resdist[i].dis)
                        currscore = currweight * res_sur[nn].score
                        weight += currweight
                        score += currscore
                        break
            if found:
                continue

            if resdist[i].dis < maxdis:
                for nn in range(len(res_lac)):
                    if res_lac[nn].nr == partner:
                        currweight = get_weight(par, resdist[i].dis)
                        currscore = currweight * res_lac[nn].score
                        weight += currweight
                        score += currscore
                        break

        res_sur[n].score = score / weight

    # Sort by scores in reverse order:
    res_sur = sorted(res_sur, key=lambda res: res.score, reverse=True)

    return res_sur


def main():

    # Parse command line
    parser = argparse.ArgumentParser(prog="parasmooth")
    parser.add_argument(
        "surface_cons_file",
        help="Surface conservation file",
        metavar="surface_cons_file",
    )
    parser.add_argument(
        "low_accessible_cons_file",
        help="Low accessible conservation file",
        metavar="low_accessible_cons_file",
    )
    parser.add_argument(
        "residue_distance_matrix",
        help="Residue distance matrix",
        metavar="residue_distance_matrix",
    )
    # parser.add_argument(
    #     "smoothing_parameter_file",
    #     help="Smoothing parameter file",
    #     metavar="smoothing_parameter_file",
    # )
    parser.add_argument(
        "-o",
        "--output",
        help="If set, output prediction to this file",
        dest="output_file",
        metavar="output_file",
    )
    args = parser.parse_args()

    logger.info("Reading input files")

    try:
        # Read .acons file
        res_sur = load_cons_file(args.surface_cons_file)

        # Read .lcons file
        res_lac = load_cons_file(args.low_accessible_cons_file)

        # Read .rd file
        resdist = load_residue_distance_matrix(args.residue_distance_matrix)

        # Read .par file
        par = load_smoothing_parameter_file(SMOOTHING_PARAM_FILE)

        # Calculate parasmooth values
        logger.info("Calculating parameter smoothing")
        res_sur = calculate_parasmooth(res_sur, res_lac, resdist, par)
    except Exception as err:
        logger.error(str(err))
        raise SystemExit

    # If output to file
    if args.output_file:
        with open(args.output_file, "w") as handle:
            for n in range(len(res_sur)):
                handle.write(
                    "{:8.5f}   {}{}{}".format(
                        res_sur[n].score, res_sur[n].code, res_sur[n].nr, os.linesep
                    )
                )
        logger.info("Result written to {}".format(args.output_file))
    else:
        for n in range(len(res_sur)):
            print(
                "{:8.5f}   {}{}".format(
                    res_sur[n].score, res_sur[n].code, res_sur[n].nr
                )
            )


if __name__ == "__main__":
    main()



"""Whiscy Data files"""

import os


class Residue:
    """Represents a residue"""

    def __init__(self, nr, code, score):
        self.nr = nr
        self.code = code
        self.score = score

    def __str__(self):
        return "{}.{}: {}".format(self.code, self.nr, self.score)


class Distance:
    """Represents the distance between two residues"""

    def __init__(self, nr1, nr2, dis):
        self.nr1 = nr1
        self.nr2 = nr2
        self.dis = dis

    def __str__(self):
        return "{} - {}: {}".format(self.nr1, self.nr2, self.dis)


_accepted_residue_codes = [
    "A",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "K",
    "L",
    "M",
    "N",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "V",
    "W",
    "Y",
]


def load_surface_list(file_name):
    """Loads the data from a surface list file.

    Tipically, this file has extension .sur and its content is a
    simple column containing the index of the residue which has been
    marked as surface.
    """
    surface_list = []
    with open(file_name, "r") as handle:
        for line in handle:
            if line:
                try:
                    surface_list.append(int(line))
                except ValueError:
                    pass
    return surface_list


def load_conversion_table(file_name):
    """Loads the data from a conversion dictionary file.

    Tipically, this file has extension .conv and its content is two columns.
    Left column is the id of the position of the residue in the original PDB
    structure. Right Column is the id of the position of the residue in Phylseq
    aligment file.
    """
    conversion_table = {}
    with open(file_name, "r") as handle:
        for line in handle:
            if line:
                fields = line.rstrip(os.linesep).split()
                try:
                    from_index, to_index = int(fields[0]), int(fields[1])
                    conversion_table[from_index] = to_index
                except ValueError:
                    pass
    return conversion_table


def load_cons_file(file_name):
    """Reads and parses a .acons file"""
    residues = []
    with open(file_name, "r") as handle:
        for line in handle:
            if line:
                fields = line.rstrip(os.linesep).split()
                try:
                    score = float(fields[0])
                    code = fields[1][0].upper()
                    nr = int(fields[1][1:])
                    residues.append(Residue(nr, code, score))
                except:
                    raise Exception(
                        "Reading error in conservation file {}".format(file_name)
                    )
    return residues


def load_residue_distance_matrix(file_name):
    """Reads and parses the residue distance matrix"""
    distances = []
    with open(file_name, "r") as handle:
        for line in handle:
            if line:
                fields = line.rstrip(os.linesep).split()
                try:
                    nr1 = int(fields[0])
                    nr2 = int(fields[1])
                    distance = float(fields[2])
                    distances.append(Distance(nr1, nr2, distance))
                except:
                    raise Exception(
                        "Reading error in distance matrix file {}".format(file_name)
                    )
    return distances


def load_smoothing_parameter_file(file_name):
    """Reads and parses the smoothing parameter file"""
    par = {}
    with open(file_name, "r") as handle:
        for line in handle:
            if line:
                fields = line.rstrip(os.linesep).split()
                try:
                    v = float(fields[0])
                    w = float(fields[1])
                    par[v] = w
                except:
                    raise Exception(
                        "Reading error in smoothing parameter file {}".format(file_name)
                    )
    return par


def load_residue_weights(file_name):
    """Loads a file containing for each line a residue in one letter code and a weight.

    Ignores lines starting with '#', only accepts a residue the first time it appears in the file
    and only if it belongs to standard list of residues.
    """
    resweight = {}
    with open(file_name, "r") as handle:
        for line in handle:
            line = line.rstrip(os.linesep)
            if line[0] != "#":
                try:
                    fields = line.split()
                    residue_id = fields[0].upper()
                    weight = float(fields[1])
                    if (
                        residue_id in _accepted_residue_codes
                        and residue_id not in resweight
                    ):
                        resweight[residue_id] = weight
                except:
                    pass
    return resweight


def load_z_table(file_name):
    """Loads the Z-table, 25000 rows"""
    z_values = []
    with open(file_name, "r") as handle:
        for line in handle:
            try:
                line = line.rstrip(os.linesep)
                z_values.append(float(line))
            except:
                pass
    return z_values



import argparse
import logging
import math
import os
import sys
from pathlib import Path

import numpy as np

from whiscy.modules import PARAM_PATH
from whiscy.modules.whiscy_data import (
    load_cons_file,
    load_residue_weights,
    load_z_table,
)

logger = logging.getLogger("consadjust")
logger.setLevel(logging.INFO)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(name)s [%(levelname)s] %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)

RESIDUE_WEIGHT_FILE = Path(PARAM_PATH, "weight_ma.txt")
ZTABLE_FILE = Path(PARAM_PATH, "ztable.txt")


def main():

    # Parse command line
    parser = argparse.ArgumentParser(prog="consadjust")
    parser.add_argument("cons_file", help="Conservation file", metavar="cons_file")

    parser.add_argument(
        "-o",
        "--output",
        help="If set, output prediction to this file",
        dest="output_file",
        metavar="output_file",
    )
    args = parser.parse_args()

    if not os.path.exists(args.cons_file):
        logger.error("Conservation file {} does not exist".format(args.cons_file))
        raise SystemExit

    logger.info("Reading input files")

    resweight = load_residue_weights(RESIDUE_WEIGHT_FILE)

    residues = load_cons_file(args.cons_file)

    # Statistics
    sum_scores = 0.0
    square_sum_scores = 0.0
    for residue in residues:
        sum_scores += residue.score
        square_sum_scores += residue.score * residue.score
    consnr = len(residues)
    stddev = math.sqrt((square_sum_scores - sum_scores * sum_scores / consnr) / consnr)
    mean = sum_scores / consnr

    if stddev <= 0.0:
        logger.error(
            "All identical values in conservation file {}".format(args.cons_file)
        )
        raise SystemExit

    zcalc = [(res.score - mean) / stddev for res in residues]

    # Load Z-matrix
    z_values = np.array(load_z_table(ZTABLE_FILE))
    if len(z_values) != 25000:
        logger.error("Reading error in Z-table {}".format(args.cons_file))
        raise SystemExit

    for n in range(consnr):
        pscore = 0.0
        neg = False
        currz = zcalc[n]
        if currz < 0:
            currz *= -1
            neg = True

        index = np.argmax(z_values < currz)
        z_values[index]

        if index >= 24999:
            pscore = 0.5
        if index <= 0:
            pscore = 0.0
        else:
            pscore = index - (currz - z_values[index]) / (
                z_values[index - 1] - z_values[index]
            )

        newneg = False
        pscore0 = pscore
        if neg:
            pscore0 = 50000 - pscore

        padj0 = 1000 * pscore0
        if resweight[residues[n].code]:
            padj0 = pscore0 / resweight[residues[n].code]

        padj = padj0
        if padj > 25000:
            padj = 50000 - padj
            newneg = True

        if padj < 0:
            padj = 0

        pmin = int(math.floor(padj))
        pmax = int(math.ceil(padj))
        fac = padj - pmin
        znorm = fac * z_values[pmin] + (1 - fac) * z_values[pmax]

        residues[n].score = mean + znorm * (1.0 - 2.0 * newneg) * stddev

    logger.info("Subtracting average value...")
    scoresum = 0.0
    for n in range(consnr):
        scoresum += residues[n].score
    scoreaverage = scoresum / consnr
    for n in range(consnr):
        residues[n].score -= scoreaverage

    logger.info("Sorting scores...")
    residues_sorted = sorted(residues, key=lambda res: res.score, reverse=True)

    logger.info("Writing scores...")
    # If output to file
    if args.output_file:
        with open(args.output_file, "w") as output_handle:
            for res in residues_sorted:
                output_handle.write(
                    "{0:7.5f}  {1}{2}{3}".format(
                        res.score, res.code, res.nr, os.linesep
                    )
                )

    else:
        for res in residues_sorted:
            print("{0:7.5f}  {1}{2}".format(res.score, res.code, res.nr))


if __name__ == "__main__":
    main()



import os
from pathlib import Path

MUSCLE_BIN = os.environ.get("MUSCLE_BIN")
FREESASA_BIN = os.environ.get("FREESASA_BIN")
HSSPCONV_BIN = os.environ.get("HSSPCONV_BIN")
PROTDIST_BIN = os.environ.get("PROTDIST_BIN")

# Make sure none of them are None
if MUSCLE_BIN is None:
    raise ValueError("MUSCLE_BIN not found in system variables")

if FREESASA_BIN is None:
    raise ValueError("FREESASA_BIN not found in system variables")

if HSSPCONV_BIN is None:
    raise ValueError("HSSPCONV_BIN not found in system variables")

if PROTDIST_BIN is None:
    raise ValueError("PROTDIST_BIN not found in system variables")

PARAM_PATH = Path(Path(__file__).parent.parent, "param")


CUTOFF = {
    "sa_pred_cutoff": 15.0,
    "sa_act_cutoff": 40.0,
    "air_cutoff": 0.18,
    "air_dist_cutoff": 6.5,
}

AIR = {
    "air_pro_percentage": 10.0,
    "air_wm_pro_or": 98.52,
    "air_wm_whis_or": 0.370515,
    "air_wm_pro_and": 55.42,
    "air_wm_whis_and": 0.106667,
}



import bz2
import logging
import os
import sys
import urllib.request
from ftplib import FTP
from typing import Union

from Bio import AlignIO

logger = logging.getLogger("whiscy_setup")


def get_from_ftp(
    pdb_code: str,
    path_to_store: str = ".",
    ftp_server: str = "ftp.cmbi.umcn.nl",
    ftp_path: str = "/pub/molbio/data/hssp/",
) -> Union[str, None]:
    """Downloads an HSSP alignment file for the given pdb_code using FTP protocol.

    Args:
        pdb_code (str): The PDB code for which to download the HSSP alignment file.
        path_to_store (str, optional): The path where the downloaded file should be stored. Defaults to ".".
        ftp_server (str, optional): The FTP server to connect to. Defaults to "ftp.cmbi.umcn.nl".
        ftp_path (str, optional): The path on the FTP server where the HSSP alignments are stored. Defaults to "/pub/molbio/data/hssp/".

    Returns:
        Union[str, None]: The path to the downloaded file if successful, None otherwise.
    """
    path_to_file: str = ""
    # Start connection
    try:
        ftp = FTP(ftp_server)
        # Anonymous login
        ftp.login()
        # Move to path where HSSP alignments are stored
        ftp.cwd(ftp_path)
        # File name format
        file_name = "{}.hssp.bz2".format(pdb_code.lower())
        # Retrieve file
        path_to_file = os.path.join(path_to_store, file_name)
        ftp.retrbinary("RETR " + file_name, open(path_to_file, "wb").write)
        # Close connection
        ftp.close()

        return path_to_file
    except Exception as e:
        logger.debug(e)
        if path_to_file != "" and os.path.exists(path_to_file):
            os.remove(path_to_file)
        return None


def get_from_url(
    pdb_code, path_to_store=".", url="ftp://ftp.cmbi.umcn.nl/pub/molbio/data/hssp3/"
):
    """Downloads from HSSP3 online db the HSSP file in stockholm format"""
    file_name = "{}.hssp.bz2".format(pdb_code.lower())
    # Make sure we use hssp3 instead of simple hssp to help identifying them
    path_to_file = os.path.join(path_to_store, file_name.replace("hssp", "hssp3"))
    urllib.request.urlretrieve(url + file_name, path_to_file)
    return path_to_file


def decompress_bz2(file_name_input, file_name_output):
    """Decompresses file_name_input in BZ2 format into file_name_output"""
    with open(file_name_output, "wb") as new_file, bz2.BZ2File(
        file_name_input, "rb"
    ) as file:
        for data in iter(lambda: file.read(100 * 1024), b""):
            new_file.write(data)


def _parse_hssp_proteins(line_buffer):
    """Parses the '## PROTEINS section' from HSSP alignment file"""
    proteins = {}
    for line in line_buffer:
        if line.startswith("  NR.") or line.startswith("##"):
            continue
        # Only get the id and name of the protein in the alignment
        fields = (line[:20]).split(":")
        seq_id = int(fields[0]) - 1
        name = fields[1].strip()[:10]
        proteins[seq_id] = name
    return proteins


def _parse_hssp_alignments(line_buffer, chain_id, num_alignments):
    """Parses the '## ALIGNMENTS section' from HSSP alignment file"""
    alignments = [[] for i in range(num_alignments)]
    first_alignment = 0
    last_alignment = 0
    current_num_alignments = 0
    for line in line_buffer:
        if line.startswith(" SeqNo") or line[12] == "!":
            continue
        if line.startswith("## ALIGNMENTS"):
            fields = (line[13:]).split("-")
            # We are now parsing alignments from first to last specified
            # in the ALINGMENTS header
            first_alignment = int(fields[0]) - 1
            last_alignment = int(fields[1]) - 1
            current_num_alignments = last_alignment - first_alignment + 1
        else:
            if line[12] == chain_id and line[14] != "X":
                for i, s in enumerate(line[51 : 51 + current_num_alignments]):
                    # We will convert spaces or dots to -
                    if s == "." or s == " ":
                        s = "-"
                    # We leave residues in minor case as if to not forget insertions
                    alignments[first_alignment + i].append(s)
    alignments = [("".join(s)) for s in alignments]
    return alignments


def hssp_file_to_phylip(hssp_file_name, phylip_file_name, chain_id, master_sequence):
    """Parses an HSSP file and returns a list of the sequences"""
    # We're only interested in the lenght of the sequence of our given chain_id,
    # SEQLENGHT header gives us the sum of all.
    seqlength = len(master_sequence)
    num_alignments = 0
    num_chains = 0
    parsing = False
    parsing_alignment = False
    line_buffer = []
    parsing_proteins = False
    prot_line_buffer = []
    with open(hssp_file_name, "r") as handle:
        for line in handle:
            line = line.rstrip(os.linesep)
            if line.startswith("NCHAIN"):
                num_chains = int(line.split()[1])
            if line.startswith("NALIGN"):
                num_alignments = int(line.split()[1])

            parsing = seqlength != 0 and num_chains != 0 and num_alignments != 0

            if parsing:
                if line.startswith("## ALIGNMENTS"):
                    parsing_alignment = True

                if line.startswith("## PROTEINS"):
                    parsing_proteins = True

                if line.startswith("##") and "ALIGNMENTS" not in line:
                    parsing_alignment = False

                if line.startswith("##") and "PROTEINS" not in line:
                    parsing_proteins = False

                if parsing_alignment:
                    line_buffer.append(line)

                if parsing_proteins:
                    prot_line_buffer.append(line)

        proteins = _parse_hssp_proteins(prot_line_buffer)
        alignments = _parse_hssp_alignments(
            line_buffer, chain_id.upper(), num_alignments
        )

        all_zero = sum([len(a) for a in alignments]) == 0

        if all_zero:
            raise Exception(
                "Not a single alignment found for chain {}".format(chain_id)
            )

        non_valid = [
            k for k in proteins.keys() if alignments[k].count("-") >= seqlength
        ]
        with open(phylip_file_name, "w") as output_handle:
            # Write header, MASTER also counts
            output_handle.write(
                "{}  {}{}".format(
                    len(proteins) - len(non_valid) + 1, seqlength, os.linesep
                )
            )
            # Write master sequence
            output_handle.write("MASTER    {}{}".format(master_sequence, os.linesep))
            # Write the rest of non null alignments
            for k in sorted(proteins.keys()):
                if k not in non_valid:
                    output_handle.write(
                        "{:10s}{}{}".format(proteins[k], alignments[k], os.linesep)
                    )


def hssp3_file_to_phylip(hssp3_file_name, phylip_file_name, chain_id, master_sequence):
    """Reads a HSSP file in stockholm format and writes a new msa file in phylip-sequential format
    only containing the given chain"""
    alignments = list(AlignIO.parse(hssp3_file_name, format="stockholm"))
    for align in alignments:
        if align[0].name[4] == "/":  # type: ignore
            chain = align[0].name[5].upper()  # type: ignore
            if chain == chain_id:
                align[0].id = align[0].name = align[0].description = "MASTER"  # type: ignore
                # align[0].seq = align[0].seq.ungap('-')
                AlignIO.write(align, phylip_file_name, format="phylip-sequential")



"""
Calculates the active/passive residues for HADDOCK

- Active residues: >=0.18 (WHISCY score)
- Passive residues: radius of 6.5A of active residues
"""

import argparse
import os

import Bio.PDB

STANDARD_TYPES = {
    "ALA": "A",
    "ARG": "R",
    "ASN": "N",
    "ASP": "D",
    "CYS": "C",
    "GLU": "E",
    "GLN": "Q",
    "GLY": "G",
    "HIS": "H",
    "ILE": "I",
    "LEU": "L",
    "LYS": "K",
    "MET": "M",
    "PHE": "F",
    "PRO": "P",
    "SER": "S",
    "THR": "T",
    "TRP": "W",
    "TYR": "Y",
    "VAL": "V",
}


def parse_whiscy_scores(file_name):
    """Parses a WHISCY scores output file"""
    scores = {}
    with open(file_name) as input:
        for line in input:
            line = line.rstrip(os.linesep)
            fields = line.split()
            res = fields[1]
            score = float(fields[0])
            scores[res] = score
    return scores


def main():

    # Parse command line
    parser = argparse.ArgumentParser(prog="whiscy2haddock")
    parser.add_argument(
        "input_pdb_file", help="Input PDB structure file", metavar="input_pdb_file"
    )
    parser.add_argument(
        "output_pdb_file", help="Output PDB file", metavar="output_pdb_file"
    )
    parser.add_argument(
        "whiscy_scores_file", help="WHISCY scores file", metavar="whiscy_scores_file"
    )
    args = parser.parse_args()

    input_pdb_file_name = args.input_pdb_file
    output_pdb_file_name = args.output_pdb_file
    whiscy_scores_file_name = args.whiscy_scores_file

    # Parse scores from the WHISCY output file
    scores = parse_whiscy_scores(whiscy_scores_file_name)

    # Parse input PDB file
    pdb_parser = Bio.PDB.PDBParser(QUIET=True)
    structure = pdb_parser.get_structure(input_pdb_file_name, input_pdb_file_name)
    residues = structure.get_residues()

    # Calculate active residues
    active_residues = []
    for residue in residues:
        try:
            res_id = "{}{}".format(
                STANDARD_TYPES[residue.get_resname()], residue.get_id()[1]
            )
            if scores[res_id] >= 0.18:
                active_residues.append(residue)
        except KeyError:
            pass

    # Calculate passive residues
    atoms = Bio.PDB.Selection.unfold_entities(structure, "A")
    neighbor_search = Bio.PDB.NeighborSearch(atoms)
    passive_residue_ids = []
    for active in active_residues:
        center = active["CA"].coord
        neighbors = neighbor_search.search(center, 6.5, level="R")
        # Remove itself
        neighbors.remove(active)
        for passive in neighbors:
            res_id = "{}{}".format(
                STANDARD_TYPES[passive.get_resname()], passive.get_id()[1]  # type: ignore
            )
            passive_residue_ids.append(res_id)

    # Create unique ids for passive residues
    passive_residue_ids = list(set(passive_residue_ids))

    # Create same ID list for active:
    active_residue_ids = []
    for residue in active_residues:
        res_id = "{}{}".format(
            STANDARD_TYPES[residue.get_resname()], residue.get_id()[1]
        )
        active_residue_ids.append(res_id)
    active_residue_ids = list(set(active_residue_ids))

    # Finally, remove from passive list the active:
    passive_residue_ids = list(set(passive_residue_ids) - set(active_residue_ids))

    with open(input_pdb_file_name, "r") as input:
        with open(output_pdb_file_name, "w") as output:
            for line in input:
                line = line.rstrip(os.linesep)
                if line and line.startswith("ATOM  "):
                    res = STANDARD_TYPES[line[17:20].strip()]
                    res_num = line[22:26].strip()
                    res_id = "{}{}".format(res, res_num)
                    try:
                        if res_id in active_residue_ids:
                            bfactor = 100.0
                        elif res_id in passive_residue_ids:
                            bfactor = -100.0
                        else:
                            bfactor = 0.0
                    except KeyError:
                        bfactor = 0.0
                    line = line[:61] + "%6.2f" % bfactor + line[66:]
                output.write(line + os.linesep)

    with open("active.txt", "w") as output:
        output.write(
            ",".join(
                [str(x) for x in sorted([int(id[1:]) for id in active_residue_ids])]
            )
        )

    with open("passive.txt", "w") as output:
        output.write(
            ",".join(
                [str(x) for x in sorted([int(id[1:]) for id in passive_residue_ids])]
            )
        )


if __name__ == "__main__":
    main()



"""This script calculates the minimum distance between all residues in the PDB structure.

It assumes that the PDB file contains only one chain and there are no residues with
alternative positions. As this script should be executed after whiscy_setup, it is safe
to make such assumptions.
"""

import argparse
import logging
import os
import sys

from Bio.PDB.PDBParser import PDBParser

from whiscy.modules.pdbutil import is_hydrogen
from whiscy.modules.whiscy_data import load_conversion_table

logger = logging.getLogger("residue_distance")
logger.setLevel(logging.INFO)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(name)s [%(levelname)s] %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)


def main():

    # Parse command line
    parser = argparse.ArgumentParser(prog="residue_distance")
    parser.add_argument("pdb_file", help="PDB file", metavar="pdb_file")
    parser.add_argument("conv_file", help="Conversion table file", metavar="conv_file")
    parser.add_argument("output_file", help="Output file name", metavar="output_file")
    args = parser.parse_args()

    # Read conversion table file
    logger.info("Reading conversion table")
    conversion_table = load_conversion_table(args.conv_file)

    # Parse PDB structure
    logger.info("Reading PDB structure from {}".format(args.pdb_file))
    pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
    structure = pdb_parser.get_structure(args.pdb_file, args.pdb_file)
    model = structure[0]
    residues = [residue for residue in model.get_residues()]

    # Calculate minimum distance between residues
    keys = conversion_table.keys()
    with open(args.output_file, "w") as output_handle:
        for i in range(len(residues)):
            res1 = residues[i]
            if res1.id[1] in keys:
                for j in range(i + 1, len(residues)):
                    min_distance = float("inf")
                    res2 = residues[j]
                    if res2.id[1] in keys:
                        for a1 in res1:
                            if not is_hydrogen(a1):
                                for a2 in res2:
                                    if not is_hydrogen(a2):
                                        d = a1 - a2
                                        if d < min_distance:
                                            min_distance = d
                        output_handle.write(
                            "{0} {1} {2:.6f}{3}".format(
                                res1.id[1], res2.id[1], min_distance, os.linesep
                            )
                        )

    logger.info("Residue distances written to {}".format(args.output_file))


if __name__ == "__main__":
    main()



import argparse
import os

STANDARD_TYPES = {
    "ALA": "A",
    "ARG": "R",
    "ASN": "N",
    "ASP": "D",
    "CYS": "C",
    "GLU": "E",
    "GLN": "Q",
    "GLY": "G",
    "HIS": "H",
    "ILE": "I",
    "LEU": "L",
    "LYS": "K",
    "MET": "M",
    "PHE": "F",
    "PRO": "P",
    "SER": "S",
    "THR": "T",
    "TRP": "W",
    "TYR": "Y",
    "VAL": "V",
}


def parse_whiscy_scores(file_name):
    """Parses a WHISCY scores output file"""
    scores = {}
    with open(file_name) as input:
        for line in input:
            line = line.rstrip(os.linesep)
            fields = line.split()
            res = fields[1]
            score = float(fields[0])
            scores[res] = score
    return scores


def main():

    # Parse command line
    parser = argparse.ArgumentParser(prog="whiscy2bfactor")
    parser.add_argument(
        "input_pdb_file", help="Input PDB structure file", metavar="input_pdb_file"
    )
    parser.add_argument(
        "output_pdb_file", help="Output PDB file", metavar="output_pdb_file"
    )
    parser.add_argument(
        "whiscy_scores_file", help="WHISCY scores file", metavar="whiscy_scores_file"
    )
    parser.add_argument(
        "--norm",
        help="Normalize WHISCY scores",
        dest="norm",
        action="store_true",
        default=False,
    )
    args = parser.parse_args()

    input_pdb_file_name = args.input_pdb_file
    output_pdb_file_name = args.output_pdb_file
    whiscy_scores_file_name = args.whiscy_scores_file

    # Parse scores from the WHISCY output file
    scores = parse_whiscy_scores(whiscy_scores_file_name)

    with open(input_pdb_file_name, "r") as input:
        with open(output_pdb_file_name, "w") as output:
            for line in input:
                line = line.rstrip(os.linesep)
                if line and line.startswith("ATOM  "):
                    try:
                        res = STANDARD_TYPES[line[17:20].strip()]
                        res_num = line[22:26].strip()
                        res_id = "{}{}".format(res, res_num)
                        score = scores[res_id]
                        if args.norm:
                            # Normalized Score = 100 * WHISCYSCORE + 50
                            # Values are truncated between 0 and 100
                            score = max(0.0, min(100.0, 100.0 * (score + 50.0)))
                    except KeyError:
                        score = 0.0

                    line = line[:61] + "%6.2f" % score + line[66:]
                output.write(line + os.linesep)

    print(
        "{} PDB file with WHISCY scores in B-factor column has been created".format(
            output_pdb_file_name
        )
    )


if __name__ == "__main__":
    main()



import random

_quotes = {
    "Mark Twain": "Too much of anything is bad, but too much good whiskey is barely enough.",
    "Raymond Chandler": "There is no bad whiskey. There are only some whiskeys that aren’t as good as others.",
    "Winston Churchill": "The water was not fit to drink. To make it palatable, we had to add whisky. By diligent effort, I learned to like it.",
    "Tommy Cooper": "I’m on a whisky diet. I’ve lost three days already.",
    "Humphrey Bogart": "I should never have switched from Scotch to Martinis.",
    "Johnny Carson": "Happiness is having a rare steak, a bottle of whisky, and a dog to eat the rare steak.",
    "Abraham Lincoln": "Tell me what brand of whiskey that Grant drinks. I would like to send a barrel of it to my other generals.",
    "Alexander Fleming": "A good gulp of hot whiskey at bedtime—it’s not very scientific, but it helps.",
    "W.C. Fields": "Always carry a flagon of whiskey in case of snakebite and furthermore always carry a small snake.",
    "James Joyce": "The light music of whiskey falling into a glass—an agreeable interlude.",
    "Igor Stravinsky": "My God, so much I like to drink Scotch that sometimes I think my name is Igor Stra-whiskey.",
    "Haruki Murakami": "The second whiskey is always my favorite. From the third on, it no longer has any taste. It's just something to pour into your stomach.",
    "Mikael Trellet": "Whiscy, to use without moderation.",
    "Jorge Roel": "Coffee keeps me going until it's acceptable to drink Whiskey",
}


def get_one():
    one = random.choice(list(_quotes))
    return '"{}"\n  -  {}'.format(_quotes[one], one)



import os
import subprocess

from whiscy.modules import FREESASA_BIN


def calculate_accessibility(pdb_file_name, output_file_name):
    """Calculates the SASA using freesasa.

    Uses the command line interface and not the Python bindings to be able to get
    a RSA NACCESS-format like file.
    """
    cmd = f"{FREESASA_BIN} {pdb_file_name} -n 20 --format=rsa --radii=naccess -o {output_file_name}"

    try:
        subprocess.run(cmd, shell=True)
    except:
        subprocess.check_call(cmd, shell=True)


class ResidueSASA:
    def __init__(self, chain_id, name, number, tot_rel, sd_rel, bk_rel):
        self.chain_id = chain_id
        self.name = name
        self.number = number
        self.tot_rel = tot_rel
        self.sd_rel = sd_rel
        self.bk_rel = bk_rel


def parse_rsa_file(rsa_file_name):
    """Parses a .rsa NACCESS (or freesasa) file and gets the relative SASAs"""
    residue_sasas = []
    with open(rsa_file_name, "r") as input_handle:
        for line in input_handle:
            if line.startswith("RES"):
                if line[13] == " ":
                    # Avoid alternative positions
                    name = line[4:7]
                    chain_id = line[8]
                    number = int(line[9:13])
                    # Difference between NACCESS and freesasa output
                    try:
                        tot_rel = float(line[23:29])
                    except ValueError:
                        tot_rel = -99.9
                    try:
                        sd_rel = float(line[36:42])
                    except ValueError:
                        sd_rel = -99.9
                    try:
                        bk_rel = float(line[49:55])
                    except ValueError:
                        bk_rel = -99.9

                    residue_sasas.append(
                        ResidueSASA(chain_id, name, number, tot_rel, sd_rel, bk_rel)
                    )

    return residue_sasas


def create_cutoff_files(rsa_file_name, pdb_code, chain_id, cutoffs, path="."):
    """Creates three output files depending on cutoffs of accessibility.

    - pdb_code.sur: rsa_file_name filtered by surface sa_pred_cutoff
    - pdb_code.suract: rsa_file_name filtered by surface sa_act_cutoff
    - pdb_code.lac: rsa_file_name filtered by buried sa_pred_cutoff
    """
    residue_sasas = parse_rsa_file(rsa_file_name)

    # Create .sur file
    output_file_name = os.path.join(path, "{0}_{1}.sur".format(pdb_code, chain_id))
    cutoff = cutoffs["sa_pred_cutoff"]
    with open(output_file_name, "w") as output_handle:
        for res_sasa in residue_sasas:
            if res_sasa.chain_id == chain_id:
                if (
                    res_sasa.tot_rel >= cutoff
                    or res_sasa.sd_rel >= cutoff
                    or res_sasa.bk_rel >= cutoff
                ):
                    output_handle.write("{}{}".format(res_sasa.number, os.linesep))

    # Create .suract file
    output_file_name = os.path.join(path, "{0}_{1}.suract".format(pdb_code, chain_id))
    cutoff = cutoffs["sa_act_cutoff"]
    with open(output_file_name, "w") as output_handle:
        for res_sasa in residue_sasas:
            if res_sasa.chain_id == chain_id:
                if (
                    res_sasa.tot_rel >= cutoff
                    or res_sasa.sd_rel >= cutoff
                    or res_sasa.bk_rel >= cutoff
                ):
                    output_handle.write("{}{}".format(res_sasa.number, os.linesep))

    # Create .lac file
    output_file_name = os.path.join(path, "{0}_{1}.lac".format(pdb_code, chain_id))
    cutoff = cutoffs["sa_pred_cutoff"]
    with open(output_file_name, "w") as output_handle:
        for res_sasa in residue_sasas:
            if res_sasa.chain_id == chain_id:
                if (
                    (res_sasa.tot_rel < cutoff and res_sasa.tot_rel > 0)
                    and res_sasa.sd_rel < cutoff
                    and res_sasa.bk_rel < cutoff
                ):
                    output_handle.write("{}{}".format(res_sasa.number, os.linesep))



#!/usr/bin/env python3

"""Python 3 implementation of the Whiscy C binary"""

__version__ = "1.0"

import argparse

# Logging
import logging
import os
import sys

from whiscy.modules.pam_calc import pam_calc_similarity, pam_load_sequences
from whiscy.modules.pam_data import code
from whiscy.modules.quotes import get_one
from whiscy.modules.whiscy_data import load_conversion_table, load_surface_list

logger = logging.getLogger("whiscy")
logger.setLevel(logging.INFO)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(name)s [%(levelname)s] %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)


class Residue:
    """Represents a residue number and its score"""

    def __init__(self, nr, score):
        self.nr = nr
        self.score = score


def main():

    # Parse command line
    parser = argparse.ArgumentParser(prog="whiscy")
    parser.add_argument("surface_list", help="Surface list", metavar="surface_list")
    parser.add_argument(
        "conversion_table", help="Conversion table", metavar="conversion_table"
    )
    parser.add_argument(
        "alignment_file", help="Alignment file", metavar="alignment_file"
    )
    parser.add_argument("distance_file", help="Distance file", metavar="distance_file")
    parser.add_argument(
        "-o",
        "--output",
        help="If set, output prediction to this file",
        dest="output_file",
        metavar="output_file",
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s {}".format(__version__)
    )
    args = parser.parse_args()

    logger.info("Parsing surface list...")
    surface_list = load_surface_list(args.surface_list)

    logger.info("Loading conversion table...")
    conversion_table = load_conversion_table(args.conversion_table)

    logger.info("Converting...")
    converted_surface = []
    for n in range(len(surface_list)):
        if (
            not surface_list[n] in conversion_table
            or conversion_table[surface_list[n]] < 1
        ):
            logger.warning(
                "Surface residue number {0} cannot be converted".format(surface_list[n])
            )
            logger.info("Continuing program...")
        else:
            converted_surface.append(conversion_table[surface_list[n]])

    if not len(converted_surface):
        logger.error("No surface residues")
        raise SystemExit

    logger.info("Initializing score calculation...")
    try:
        seqnr, seqlen, refseq, seq_distances, sequences, seqtodis = pam_load_sequences(
            args.alignment_file, args.distance_file
        )
    except Exception as err:
        logger.error(str(err))
        raise SystemExit

    logger.info("Calculating scores...")
    realsum = 0
    totlist = []
    for n in range(len(converted_surface)):
        r = Residue(converted_surface[n], -1000.0)
        totlist.append(r)
        if r.nr > seqlen or r.nr < 1:
            logger.error("Surface residue out of range")
            raise SystemExit
        posnr, distances, scores = pam_calc_similarity(
            converted_surface[n] - 1, seqnr, sequences, seq_distances
        )
        if posnr <= 0 or posnr > seqnr:
            continue
        realsum += 1
        r.score = scores[posnr - 1]

    if realsum == 0:
        logger.error("No sequence information for any surface residues")
        raise SystemExit

    logger.info("Subtracting average value ...")

    scoresum = 0.0
    for n in range(realsum):
        scoresum += totlist[n].score

    scoreaverage = scoresum / realsum
    for n in range(realsum):
        totlist[n].score -= scoreaverage

    logger.info("Sorting scores...")
    sorted_totlist = sorted(totlist, key=lambda res: res.score, reverse=True)

    logger.info("Writing scores...")
    prediction = ""
    inv_conversion_table = {v: k for k, v in conversion_table.items()}
    for n in range(realsum):
        r = sorted_totlist[n]
        rid = "{0}{1}".format(refseq[r.nr - 1], inv_conversion_table[r.nr])
        prediction += "{:7.5f}  {:^5s}{}".format(r.score, rid, os.linesep)

    if args.output_file:
        with open(args.output_file, "w") as output:
            output.write(prediction)
        logger.info("Prediction written to {0}".format(args.output_file))
    else:
        print(prediction)

    # Ending with a quote
    print()
    print(get_one())


if __name__ == "__main__":
    main()



import argparse
import os
import shutil
import subprocess
import sys
import warnings

# Import SearchIO and suppress experimental warning
from Bio import AlignIO, BiopythonExperimentalWarning, BiopythonWarning, SeqIO
from Bio.Align import MultipleSeqAlignment
from Bio.Align.Applications import MuscleCommandline
from Bio.Blast import NCBIWWW
from Bio.PDB import PDBIO
from Bio.PDB.PDBParser import PDBParser

warnings.simplefilter("ignore", BiopythonWarning)
with warnings.catch_warnings():
    warnings.simplefilter("ignore", BiopythonExperimentalWarning)
    from Bio import SearchIO

# Logging
import logging

from whiscy.modules import (
    CUTOFF,
    HSSPCONV_BIN,
    MUSCLE_BIN,
    PROTDIST_BIN,
    access,
    hssp,
    pdbutil,
)

logger = logging.getLogger("whiscy_setup")
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.INFO)
formatter = logging.Formatter(
    "%(asctime)s %(module)s:%(lineno)d %(levelname)s - %(message)s"
)
ch.setFormatter(formatter)
logger.addHandler(ch)


def muscle_msa(
    input_sequence_file: str, output_alignment_file: str
) -> MultipleSeqAlignment:
    """Calculates a MSA using MUSCLE's Biopython wrapper"""

    assert MUSCLE_BIN is not None, "MUSCLE_BIN not found in system variables"

    muscle_cline = MuscleCommandline(
        MUSCLE_BIN, input=input_sequence_file, out=output_alignment_file
    )
    _, _ = muscle_cline()
    msa = AlignIO.read(output_alignment_file, "fasta")
    return msa


def ncbi_blast(fasta_file: str, output_file: str) -> None:
    """Performs a remote BLAST against the NCBI server"""
    record = SeqIO.read(fasta_file, format="fasta")
    result_handle = NCBIWWW.qblast("blastp", "nr", record.format("fasta"))
    with open(output_file, "w") as out_handle:
        out_handle.write(result_handle.read())
    result_handle.close()


def msa_to_phylseq(
    msa: MultipleSeqAlignment, master_sequence: str, output_file: str
) -> None:
    """Converts a MSA to a Phylip Seq file"""
    with open(output_file, "w") as output_handle:
        # Write header
        output_handle.write(f"{len(msa)}  {len(master_sequence)}{os.linesep}")
        # Write master sequence
        output_handle.write(f"MASTER    {master_sequence}{os.linesep}")
        # Write the rest of alignments
        for alignment in msa:
            output_handle.write(f"{alignment.id[:10]:10s}{alignment.seq}{os.linesep}")


def hsspconv(hssp_file, converted_hssp_file):
    """Converts a HSSP in version 3 to original format"""
    cmd = f"{HSSPCONV_BIN} < {hssp_file} > {converted_hssp_file}"
    try:
        subprocess.run(cmd, shell=True)
    except:
        subprocess.check_call(cmd, shell=True)


def calculate_protdist(phylip_file, protdist_output_file):
    """Calculates the protdist of the given MSA"""
    cmd = f"{PROTDIST_BIN} {phylip_file} {protdist_output_file} > /dev/null 2>&1"
    try:
        subprocess.run(cmd, shell=True)
    except:
        subprocess.check_call(cmd, shell=True)


def write_to_fasta(output_fasta_file, sequence):
    """Writes a sequence to a FASTA format file"""
    with open(output_fasta_file, "w") as output_handle:
        output_handle.write(">{0}{1}".format(output_fasta_file, os.linesep))
        n = 60
        seq = [sequence[i : i + n] for i in range(0, len(sequence), n)]
        for chunk in seq:
            output_handle.write("{0}{1}".format(chunk, os.linesep))


def main():

    # Parse command line
    parser = argparse.ArgumentParser(prog="whiscy_setup")
    parser.add_argument(
        "pdb_file_name",
        help="PDB file name (.pdb extension) or PDB code",
        metavar="pdb_file_name",
    )
    parser.add_argument("chain_id", help="Chain ID to be predicted", metavar="chain_id")
    parser.add_argument(
        "--hssp", help="HSSP ID code", dest="hssp", metavar="hssp", default=None
    )
    parser.add_argument(
        "--alignment",
        help="Alignment file",
        dest="alignment",
        metavar="alignment",
        default=None,
    )
    parser.add_argument(
        "--alignment_format",
        help="Alignment format",
        dest="alignment_format",
        metavar="alignment_format",
        default="fasta",
    )

    args = parser.parse_args()

    filename, file_extension = os.path.splitext(os.path.basename(args.pdb_file_name))
    with_pdb_code = file_extension == ""
    input_pdb_file = None

    if with_pdb_code:
        # PDB code has been specified instead of a PDB file name
        pdb_code = args.pdb_file_name
        input_pdb_file = "{0}.pdb".format(pdb_code)
        if not os.path.exists(input_pdb_file):
            try:
                pdbutil.download_pdb_structure(pdb_code, input_pdb_file)
            except Exception as err:
                logger.error(str(err))
                raise SystemExit
        else:
            logger.warning(
                "PDB structure already exists ({0}), no need to download it again".format(
                    input_pdb_file
                )
            )
    else:
        pdb_code = filename
        input_pdb_file = args.pdb_file_name

    if not os.path.exists(input_pdb_file):
        logger.error("PDB structure file {0} not found".format(input_pdb_file))
        raise SystemExit

    # Check if chain belongs to this PDB
    pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
    structure = pdb_parser.get_structure(filename, input_pdb_file)
    chain_ids = [chain.id for chain in structure.get_chains()]
    chain_id = args.chain_id.upper()
    if len(chain_id) > 1:
        logger.error("Wrong chain id {0}".format(chain_id))
        raise SystemExit
    if chain_id not in chain_ids:
        logger.error(
            "Chain {0} provided not in available chains: {1}".format(
                chain_id, str(chain_ids)
            )
        )
        raise SystemExit

    # Save only the given chain and discard residues with alternative positions
    io = PDBIO()
    current_pdb_file = "{0}_{1}.pdb".format(pdb_code, chain_id)
    for chain in structure.get_chains():
        if chain.id == chain_id:
            io.set_structure(chain)
            io.save(current_pdb_file, select=pdbutil.NotAlternative())
    logger.info(
        "PDB structure with chain {0} saved to {1}".format(chain_id, current_pdb_file)
    )

    # Calculate SASA:
    rsa_output_file = "{0}_{1}.rsa".format(pdb_code, chain_id)
    access.calculate_accessibility(current_pdb_file, rsa_output_file)
    logger.info("Atom accessibility calculated to {0}".format(rsa_output_file))

    # Calculate the different accessibility files according to the cutoffs:
    access.create_cutoff_files(rsa_output_file, pdb_code, chain_id, CUTOFF)
    logger.info("Surface and buried residues calculated")

    # Get structure sequence
    master_sequence = pdbutil.get_pdb_sequence(input_pdb_file, chain_id)
    write_to_fasta("{0}_{1}.fasta".format(filename, chain_id), master_sequence)

    if args.hssp:
        hssp_file = "{0}.hssp".format(args.hssp)
        phylip_file = "{0}_{1}.phylseq".format(args.hssp, chain_id)
    else:
        hssp_file = "{0}.hssp".format(pdb_code)
        phylip_file = "{0}_{1}.phylseq".format(pdb_code, chain_id)

    if with_pdb_code or args.hssp:
        # Get the HSSP alignment from FTP if pdb code specified
        if not os.path.exists(hssp_file):
            logger.info("Downloading HSSP alignment...")
            try:
                compressed_hssp_file = hssp.get_from_ftp(pdb_code)
                if not compressed_hssp_file:
                    compressed_hssp_file = hssp.get_from_url(pdb_code)

                if "hssp3" in compressed_hssp_file:
                    hssp_file = hssp_file.replace("hssp", "hssp3")
                hssp.decompress_bz2(compressed_hssp_file, hssp_file)
                logger.info("HSSP alignment stored to {0}".format(hssp_file))
            except Exception:
                logger.warning("HSSP file could not be downloaded")
        try:
            if "hssp3" in hssp_file:
                # HSSP downloaded file is in new HSSP3 format, need to be
                # converted back to original HSSP format using hsspconv
                converted_hssp_file = hssp_file.replace("hssp3", "hssp")
                hsspconv(hssp_file, converted_hssp_file)
                hssp_file = converted_hssp_file

            hssp.hssp_file_to_phylip(hssp_file, phylip_file, chain_id, master_sequence)
            logger.info("HSSP file converted to PHYLIP format")

        except Exception as e:
            logger.debug(e)

    if not os.path.exists(hssp_file):
        logger.info("HSSP file not found, fallback to generating MSA with blastp")
        # if no alignment is provided, will do it automatically
        if not args.alignment:
            # Run BLASTP if needed
            blast_output_file = "{0}_{1}_blast.xml".format(filename, chain_id)
            input_sequence_file = "{0}_{1}.fasta".format(filename, chain_id)
            if not os.path.exists(blast_output_file):
                logger.info("Running blastp via NCBI")
                ncbi_blast(input_sequence_file, blast_output_file)
                logger.info("Result stored in {0}".format(blast_output_file))

            assert os.path.exists(blast_output_file), "BLAST file not found"

            # Convert file to FASTA format
            blast_qresult = SearchIO.read(blast_output_file, "blast-xml")
            records = []
            for hit in blast_qresult:
                records.append(hit[0].hit)
            blast_fasta_file = "{0}_{1}_blast.fasta".format(filename, chain_id)
            SeqIO.write(records, blast_fasta_file, "fasta")

            # Multiple sequence alignment
            logger.info("Generating MSA using MUSCLE...")
            output_alignment_file = "{0}_{1}_msa.fasta".format(filename, chain_id)
            msa = muscle_msa(blast_fasta_file, output_alignment_file)
            logger.info("Done.")

            # Convert MSA to Phylipseq
            logger.info("Converting MSA file to Phylseq format...")
            output_phylseq_file = "{0}_{1}.phylseq".format(filename, chain_id)
            msa_to_phylseq(
                msa=msa,
                master_sequence=str(master_sequence),
                output_file=output_phylseq_file,
            )

            assert os.path.exists(output_phylseq_file), "PHYLIP file not found"
            logger.info(f"{output_phylseq_file} file written")

        else:
            # Alignment is provided, need to convert it to phylipseq
            # which is the format WHISCY is expecting
            alignment_file_name = args.alignment
            alignment_format = args.alignment_format.lower()
            phylip_file = "{0}_{1}.phylseq".format(pdb_code, chain_id)
            if alignment_format == "phylip":
                # Nothing to convert, just rename file if needed
                if os.path.basename(alignment_file_name) != phylip_file:
                    shutil.copyfile(alignment_file_name, phylip_file)
            else:
                AlignIO.convert(
                    alignment_file_name,
                    alignment_format,
                    phylip_file,
                    "phylip-sequential",
                )

    if not os.path.exists(phylip_file):
        logger.error("PHYLIP sequence file {0} not found".format(phylip_file))
        raise SystemExit

    # Calculate protdist
    protdist_output_file = "{0}_{1}.out".format(filename, chain_id)
    calculate_protdist(phylip_file, protdist_output_file)
    logger.info("Protdist calculated")

    # Generate conversion table file
    conv_output_file = "{0}_{1}.conv".format(filename, chain_id)
    pdbutil.map_protein_to_sequence_alignment(
        current_pdb_file, chain_id, master_sequence, phylip_file, conv_output_file
    )
    logger.info("Conversion table file generated")

    logger.info("Whiscy setup finished")


if __name__ == "__main__":
    main()



"""Util functions involving structure and PDB files"""

import os
import re

from Bio import AlignIO
from Bio.PDB import PDBList
from Bio.PDB.PDBIO import Select
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.Polypeptide import is_aa, three_to_one


class NotAlternative(Select):
    """Removes alternative AAs"""

    def accept_residue(self, residue):
        return is_aa(residue) and residue.id[2] == " "


_hydrogen = re.compile("[123 ]*H.*")


def is_hydrogen(atom):
    """Checks if atom is an hydrogen"""
    name = atom.get_id()
    return _hydrogen.match(name)


def download_pdb_structure(pdb_code, pdb_file_name, file_path="."):
    """Downloads a PDB structure from the Protein Data Bank"""
    pdbl = PDBList()
    file_name = pdbl.retrieve_pdb_file(
        pdb_code, file_format="pdb", pdir=file_path, overwrite=True
    )
    if os.path.exists(file_name):
        os.rename(file_name, pdb_file_name)
    else:
        raise Exception("Can not download structure: {0}".format(pdb_code))


def get_pdb_sequence(input_pdb_file, chain_id, mapping_output=False, with_gaps=False):
    """Gets the PDB sequence in a dictionary"""
    mapping = {}
    pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
    structure = pdb_parser.get_structure(input_pdb_file, input_pdb_file)
    model = structure[0]
    chain = model[chain_id]
    residues = list(chain)
    for res in residues:
        # Remove alternative location residues
        if "CA" in res.child_dict and is_aa(res) and res.id[2] == " ":
            try:
                mapping[res.id[1]] = three_to_one(res.get_resname())
            except KeyError:
                # Ignore non standard residues such as HIC, MSE, etc.
                pass

    if with_gaps:
        # Add missing gap residues by their residue number
        res_numbers = sorted(mapping.keys())
        start, end = res_numbers[0], res_numbers[-1]
        missing = sorted(set(range(start, end + 1)).difference(res_numbers))
        for m in missing:
            mapping[m] = "-"

    if mapping_output:
        return mapping
    else:
        return "".join([mapping[k] for k in sorted(mapping.keys())])


def map_protein_to_sequence_alignment(
    pdb_file, chain_id, sequence, phylip_file, output_file_name
):
    """Creates a dictionary .conv file mapping protein residue numeration to aligment"""
    mapping = get_pdb_sequence(pdb_file, chain_id, mapping_output=True)
    # Check if sequence is the same
    pdb_seq = "".join([mapping[k] for k in sorted(mapping.keys())])
    if pdb_seq != sequence:
        raise SystemExit("ERROR: PDB sequence doest not match sequence alignment")

    # Account for gaps in phylipseq file
    # alignment = list(AlignIO.parse(phylip_file, format='phylip-sequential'))[0]
    # master_phylip = alignment[0].seq
    # if str(master_phylip.ungap('-')) != pdb_seq:
    #    raise SystemExit("ERROR: PDB sequence doest not match sequence alignment in phylip file")

    with open(output_file_name, "w") as output_handle:
        output_handle.write(
            "# Conversion table from {} and chain {} to sequence{}".format(
                pdb_file, chain_id, os.linesep
            )
        )
        seq_res_id = 1
        for pdb_res_id in sorted(mapping.keys()):
            # Do not map gaps if any
            if mapping[pdb_res_id] != "-":
                output_handle.write(
                    "{0}     {1}{2}".format(pdb_res_id, seq_res_id, os.linesep)
                )
            seq_res_id += 1
